@@ -1,12 +1,16 @@
-# Copyright (C) 2019-2020 The Software Heritage developers
+# Copyright (C) 2019-2022 The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
 
+from concurrent.futures import ThreadPoolExecutor
+from concurrent.futures import wait as futures_wait
 import logging
 from time import time
+from traceback import format_tb
 from typing import Callable, Dict, List, Optional
 
+from humanize import naturaldelta, naturalsize
 import msgpack
 from sentry_sdk import capture_exception, push_scope
 
@@ -26,7 +30,9 @@ from tenacity.retry import retry_base
 from swh.core.statsd import statsd
 from swh.model.hashutil import hash_to_hex
 from swh.model.model import SHA1_SIZE
-from swh.objstorage.objstorage import ID_HASH_ALGO, ObjNotFoundError, ObjStorage
+from swh.objstorage.constants import ID_HASH_ALGO
+from swh.objstorage.exc import ObjNotFoundError
+from swh.objstorage.objstorage import ObjStorage
 
 logger = logging.getLogger(__name__)
 REPORTER = None
@@ -93,7 +99,11 @@ class ReplayError(Exception):
         self.exc = exc
 
     def __str__(self):
-        return "ReplayError(%s, %s)" % (self.obj_id, self.exc)
+        return "ReplayError(%s, %r, %s)" % (
+            self.obj_id,
+            self.exc,
+            format_tb(self.exc.__traceback__),
+        )
 
 
 def log_replay_retry(retry_state, sleep=None, last_result=None):
@@ -118,13 +128,13 @@ def log_replay_error(retry_state):
     error_context = {
         "obj_id": exc.obj_id,
         "operation": retry_state.fn.__name__,
-        "exc": str(exc.exc),
+        "exc": str(exc),
         "retries": retry_state.attempt_number,
     }
 
     logger.error(
         "Failed operation %(operation)s on %(obj_id)s after %(retries)s"
-        " retries: %(exc)s",
+        " retries; last exception: %(exc)s",
         error_context,
     )
 
@@ -134,7 +144,7 @@ def log_replay_error(retry_state):
         msg = msgpack.dumps(error_context)
         REPORTER(oid, msg)
 
-    return None
+    raise exc
 
 
 CONTENT_REPLAY_RETRIES = 3
@@ -184,10 +194,16 @@ def get_object(objstorage, obj_id):
 @content_replay_retry
 def put_object(objstorage, obj_id, obj):
     try:
+        logger.debug("putting %(obj_id)s", {"obj_id": hash_to_hex(obj_id)})
         with statsd.timed(CONTENT_DURATION_METRIC, tags={"request": "put"}):
-            obj = objstorage.add(obj, obj_id, check_presence=False)
+            logger.debug("storing %(obj_id)s", {"obj_id": hash_to_hex(obj_id)})
+            objstorage.add(obj, obj_id, check_presence=False)
             logger.debug("stored %(obj_id)s", {"obj_id": hash_to_hex(obj_id)})
     except Exception as exc:
+        logger.error(
+            "putting %(obj_id)s failed: %(exc)r",
+            {"obj_id": hash_to_hex(obj_id), "exc": exc},
+        )
         raise ReplayError(obj_id=obj_id, exc=exc) from None
 
 
@@ -216,6 +232,7 @@ def process_replay_objects_content(
     dst: ObjStorage,
     exclude_fn: Optional[Callable[[dict], bool]] = None,
     check_dst: bool = True,
+    concurrency: int = 16,
 ):
     """
     Takes a list of records from Kafka (see
@@ -238,11 +255,16 @@ def process_replay_objects_content(
 
     Example:
 
+    >>> import hashlib
     >>> from swh.objstorage.factory import get_objstorage
     >>> src = get_objstorage('memory')
     >>> dst = get_objstorage('memory')
-    >>> id1 = src.add(b'foo bar')
-    >>> id2 = src.add(b'baz qux')
+    >>> cnt1 = b'foo bar'
+    >>> cnt2 = b'baz qux'
+    >>> id1 = hashlib.sha1(cnt1).digest()
+    >>> id2 = hashlib.sha1(cnt2).digest()
+    >>> src.add(b'foo bar', obj_id=id1)
+    >>> src.add(b'baz qux', obj_id=id2)
     >>> kafka_partitions = {
     ...     'content': [
     ...         {
@@ -263,68 +285,101 @@ def process_replay_objects_content(
     >>> id2 in dst
     True
     """
-    vol = []
-    nb_skipped = 0
-    nb_failures = 0
+
+    def _copy_object(obj):
+        obj_id = obj[ID_HASH_ALGO]
+        logger.debug("Starting copy object %s", hash_to_hex(obj_id))
+        if obj["status"] != "visible":
+            logger.debug("skipped %s (status=%s)", hash_to_hex(obj_id), obj["status"])
+            statsd.increment(
+                CONTENT_OPERATIONS_METRIC,
+                tags={"decision": "skipped", "status": obj["status"]},
+            )
+            return "skipped", None
+        elif exclude_fn and exclude_fn(obj):
+            logger.debug("skipped %s (manually excluded)", hash_to_hex(obj_id))
+            statsd.increment(CONTENT_OPERATIONS_METRIC, tags={"decision": "excluded"})
+            return "excluded", None
+        elif check_dst and obj_in_objstorage(obj_id, dst):
+            logger.debug("skipped %s (in dst)", hash_to_hex(obj_id))
+            statsd.increment(CONTENT_OPERATIONS_METRIC, tags={"decision": "in_dst"})
+            return "in_dst", None
+        else:
+            try:
+                copied_bytes = copy_object(obj_id, src, dst)
+            except ObjNotFoundError:
+                logger.debug("not found %s", hash_to_hex(obj_id))
+                statsd.increment(
+                    CONTENT_OPERATIONS_METRIC, tags={"decision": "not_in_src"}
+                )
+                return "not_found", None
+            except Exception as exc:
+                logger.info("failed %s (%r)", hash_to_hex(obj_id), exc)
+                statsd.increment(CONTENT_OPERATIONS_METRIC, tags={"decision": "failed"})
+                return "failed", None
+            else:
+                if copied_bytes is None:
+                    logger.debug("failed %s (None)", hash_to_hex(obj_id))
+                    statsd.increment(
+                        CONTENT_OPERATIONS_METRIC, tags={"decision": "failed"}
+                    )
+                    return "failed", None
+                else:
+                    logger.debug("copied %s (%d)", hash_to_hex(obj_id), copied_bytes)
+                    statsd.increment(
+                        CONTENT_OPERATIONS_METRIC, tags={"decision": "copied"}
+                    )
+                    return "copied", copied_bytes
+        logger.debug("failed %s (XXX)", hash_to_hex(obj_id))
+        return "failed", None
+
+    vol = 0
+    stats = dict.fromkeys(
+        ["skipped", "excluded", "not_found", "failed", "copied", "in_dst"], 0
+    )
     t0 = time()
 
-    for (object_type, objects) in all_objects.items():
-        if object_type != "content":
-            logger.warning(
-                "Received a series of %s, this should not happen", object_type
-            )
-            continue
-        for obj in objects:
-            obj_id = obj[ID_HASH_ALGO]
-            if obj["status"] != "visible":
-                nb_skipped += 1
-                logger.debug(
-                    "skipped %s (status=%s)", hash_to_hex(obj_id), obj["status"]
+    with ThreadPoolExecutor(max_workers=concurrency) as pool:
+        futures = []
+        for (object_type, objects) in all_objects.items():
+            if object_type != "content":
+                logger.warning(
+                    "Received a series of %s, this should not happen", object_type
                 )
-                statsd.increment(
-                    CONTENT_OPERATIONS_METRIC,
-                    tags={"decision": "skipped", "status": obj["status"]},
-                )
-            elif exclude_fn and exclude_fn(obj):
-                nb_skipped += 1
-                logger.debug("skipped %s (manually excluded)", hash_to_hex(obj_id))
-                statsd.increment(
-                    CONTENT_OPERATIONS_METRIC, tags={"decision": "excluded"}
-                )
-            elif check_dst and obj_in_objstorage(obj_id, dst):
-                nb_skipped += 1
-                logger.debug("skipped %s (in dst)", hash_to_hex(obj_id))
-                statsd.increment(CONTENT_OPERATIONS_METRIC, tags={"decision": "in_dst"})
-            else:
-                try:
-                    copied = copy_object(obj_id, src, dst)
-                except ObjNotFoundError:
-                    nb_skipped += 1
-                    statsd.increment(
-                        CONTENT_OPERATIONS_METRIC, tags={"decision": "not_in_src"}
-                    )
-                else:
-                    if copied is None:
-                        nb_failures += 1
-                        statsd.increment(
-                            CONTENT_OPERATIONS_METRIC, tags={"decision": "failed"}
-                        )
-                    else:
-                        vol.append(copied)
-                        statsd.increment(
-                            CONTENT_OPERATIONS_METRIC, tags={"decision": "copied"}
-                        )
+                continue
+            for obj in objects:
+                futures.append(pool.submit(_copy_object, obj=obj))
+        logger.debug("Waiting for futures (%d)", len(futures))
+        futures_wait(futures)
+    logger.debug("Checking futures results")
+    for f in futures:
+        exc = f.exception()
+        if exc:
+            # XXX this should not happen, so it is probably wrong...
+            raise exc
+        else:
+            state, nbytes = f.result()
+            if nbytes is not None:
+                vol += nbytes
+            stats[state] += 1
 
     dt = time() - t0
     logger.info(
-        "processed %s content objects in %.1fsec "
-        "(%.1f obj/sec, %.1fMB/sec) - %d failed - %d skipped",
-        len(vol),
-        dt,
-        len(vol) / dt,
-        sum(vol) / 1024 / 1024 / dt,
-        nb_failures,
-        nb_skipped,
+        "processed %s content objects (%s) in %s "
+        "(%.1f obj/sec, %s/sec) "
+        "- %d copied - %d in dst - %d skipped "
+        "- %d excluded - %d not found - %d failed",
+        len(futures),
+        naturalsize(vol),
+        naturaldelta(dt),
+        len(futures) / dt,
+        naturalsize(vol / dt),
+        stats["copied"],
+        stats["in_dst"],
+        stats["skipped"],
+        stats["excluded"],
+        stats["not_found"],
+        stats["failed"],
     )
 
     if notify:
