@@ -1,4 +1,4 @@
-# Copyright (C) 2019-2020  The Software Heritage developers
+# Copyright (C) 2019-2022  The Software Heritage developers
 # See the AUTHORS file at the top-level directory of this distribution
 # License: GNU General Public License version 3, or any later version
 # See top-level LICENSE file for more information
@@ -10,7 +10,7 @@ import logging
 import re
 from subprocess import Popen
 import tempfile
-from typing import Tuple
+from typing import Any, Dict, Optional, Tuple
 from unittest.mock import patch
 
 from click.testing import CliRunner
@@ -20,7 +20,7 @@ import pytest
 import yaml
 
 from swh.journal.serializers import key_to_kafka
-from swh.model.hashutil import hash_to_hex
+from swh.model.hashutil import MultiHash, hash_to_hex
 from swh.objstorage.backends.in_memory import InMemoryObjStorage
 from swh.objstorage.replayer.cli import objstorage_cli_group
 from swh.objstorage.replayer.replay import CONTENT_REPLAY_RETRIES
@@ -29,8 +29,14 @@ logger = logging.getLogger(__name__)
 
 
 CLI_CONFIG = {
-    "objstorage": {"cls": "mocked", "name": "src",},
-    "objstorage_dst": {"cls": "mocked", "name": "dst",},
+    "objstorage": {
+        "cls": "mocked",
+        "name": "src",
+    },
+    "objstorage_dst": {
+        "cls": "mocked",
+        "name": "dst",
+    },
 }
 
 
@@ -62,10 +68,9 @@ def _patch_objstorages(names):
     return decorator
 
 
-def invoke(*args, env=None, journal_config=None):
+def invoke(*args, env=None, **kwargs):
     config = copy.deepcopy(CLI_CONFIG)
-    if journal_config:
-        config["journal_client"] = journal_config
+    config.update(kwargs)
 
     runner = CliRunner()
     with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
@@ -73,12 +78,18 @@ def invoke(*args, env=None, journal_config=None):
         config_fd.seek(0)
         args = ["-C" + config_fd.name] + list(args)
         return runner.invoke(
-            objstorage_cli_group, args, obj={"log_level": logging.DEBUG}, env=env,
+            objstorage_cli_group,
+            args,
+            obj={"log_level": logging.DEBUG},
+            env=env,
         )
 
 
 def test_replay_help():
-    result = invoke("replay", "--help",)
+    result = invoke(
+        "replay",
+        "--help",
+    )
     expected = (
         r"^\s*Usage: objstorage replay \[OPTIONS\]\s+"
         r"Fill a destination Object Storage.*"
@@ -102,12 +113,18 @@ def _fill_objstorage_and_kafka(kafka_server, kafka_prefix, objstorage):
     contents = {}
     for i in range(NUM_CONTENTS):
         content = b"\x00" * 19 + bytes([i])
-        sha1 = objstorage.add(content)
+        sha1 = MultiHash(["sha1"]).from_data(content).digest()["sha1"]
+        objstorage.add(content=content, obj_id=sha1)
         contents[sha1] = content
         producer.produce(
             topic=kafka_prefix + ".content",
             key=key_to_kafka(sha1),
-            value=key_to_kafka({"sha1": sha1, "status": "visible",}),
+            value=key_to_kafka(
+                {
+                    "sha1": sha1,
+                    "status": "visible",
+                }
+            ),
         )
 
     producer.flush()
@@ -132,7 +149,8 @@ def test_replay_content(
         "replay",
         "--stop-after-objects",
         str(NUM_CONTENTS),
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
@@ -170,7 +188,8 @@ def test_replay_content_structured_log(
         "replay",
         "--stop-after-objects",
         str(NUM_CONTENTS),
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
@@ -216,7 +235,8 @@ def test_replay_content_static_group_id(
         "--stop-after-objects",
         str(NUM_CONTENTS),
         env={"KAFKA_GROUP_INSTANCE_ID": "static-group-instance-id"},
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
@@ -226,16 +246,19 @@ def test_replay_content_static_group_id(
     assert result.exit_code == 0, result.output
     assert re.fullmatch(expected, result.output, re.MULTILINE), result.output
 
-    consumer_settings = None
+    consumer_settings: Optional[Dict[str, Any]] = None
     for record in caplog.records:
         if "Consumer settings" in record.message:
-            consumer_settings = record.args
-            break
+            consumer_settings = {}
+        elif consumer_settings is not None and len(record.args) == 2:
+            key, value = record.args
+            consumer_settings[key] = value
 
     assert consumer_settings is not None, (
         "Failed to get consumer settings out of the consumer log. "
         "See log capture for details."
     )
+
     assert consumer_settings["group.instance.id"] == "static-group-instance-id"
     assert consumer_settings["session.timeout.ms"] == 60 * 10 * 1000
     assert consumer_settings["max.poll.interval.ms"] == 90 * 10 * 1000
@@ -273,7 +296,8 @@ def test_replay_content_exclude(
             str(NUM_CONTENTS),
             "--exclude-sha1-file",
             fd.name,
-            journal_config={
+            journal_client={
+                "cls": "kafka",
                 "brokers": kafka_server,
                 "group_id": kafka_consumer_group,
                 "prefix": kafka_prefix,
@@ -337,7 +361,8 @@ def test_replay_content_check_dst(
         "--stop-after-objects",
         str(NUM_CONTENTS),
         "--check-dst" if check_dst else "--no-check-dst",
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
@@ -347,22 +372,29 @@ def test_replay_content_check_dst(
     assert result.exit_code == 0, result.output
     assert re.fullmatch(expected, result.output, re.MULTILINE), result.output
 
-    retrieved = 0
-    stored = 0
-    in_dst = 0
+    stats = dict.fromkeys(
+        ["tot", "copied", "in_dst", "skipped", "excluded", "not_found", "failed"], 0
+    )
+    reg = re.compile(
+        r"processed (?P<tot>\d+) content objects .*"
+        r" *- (?P<copied>\d+) copied"
+        r" *- (?P<in_dst>\d+) in dst"
+        r" *- (?P<skipped>\d+) skipped"
+        r" *- (?P<excluded>\d+) excluded"
+        r" *- (?P<not_found>\d+) not found"
+        r" *- (?P<failed>\d+) failed"
+    )
     for record in caplog.records:
         logtext = record.getMessage()
-        if "retrieved" in logtext:
-            retrieved += 1
-        elif "stored" in logtext:
-            stored += 1
-        elif "in dst" in logtext:
-            in_dst += 1
+        m = reg.match(logtext)
+        if m:
+            for k, v in m.groupdict().items():
+                stats[k] += int(v)
+
+    assert stats["tot"] == sum(v for k, v in stats.items() if k != "tot")
 
     assert (
-        retrieved == expected_copied
-        and stored == expected_copied
-        and in_dst == expected_in_dst
+        stats["copied"] == expected_copied and stats["in_dst"] == expected_in_dst
     ), "Unexpected amount of objects copied, see the captured log for details"
 
     for (sha1, content) in contents.items():
@@ -446,10 +478,13 @@ def test_replay_content_check_dst_retry(
         "--check-dst",
         "--stop-after-objects",
         str(NUM_CONTENTS),
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
+        },
+        replayer={
             "error_reporter": {"host": redis_proc.host, "port": redis_proc.port},
         },
     )
@@ -532,10 +567,12 @@ def test_replay_content_failed_copy_retry(
     assert get_failures
     assert definitely_failed
     objstorages["dst"] = FlakyObjStorage(
-        state=objstorages["dst"].state, failures=add_failures,
+        state=objstorages["dst"].state,
+        failures=add_failures,
     )
     objstorages["src"] = FlakyObjStorage(
-        state=objstorages["src"].state, failures=get_failures,
+        state=objstorages["src"].state,
+        failures=get_failures,
     )
 
     caplog.set_level(logging.DEBUG, "swh.objstorage.replayer.replay")
@@ -544,10 +581,13 @@ def test_replay_content_failed_copy_retry(
         "replay",
         "--stop-after-objects",
         str(NUM_CONTENTS),
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
+        },
+        replayer={
             "error_reporter": {"host": redis_proc.host, "port": redis_proc.port},
         },
     )
@@ -628,7 +668,8 @@ def test_replay_content_objnotfound(
         "replay",
         "--stop-after-objects",
         str(NUM_CONTENTS),
-        journal_config={
+        journal_client={
+            "cls": "kafka",
             "brokers": kafka_server,
             "group_id": kafka_consumer_group,
             "prefix": kafka_prefix,
